﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Data;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Random;

namespace HeuristicLab.VariableInteractionNetworks {
  /// <summary>
  /// For the variable interaction network we found the existing code for variable impact calculation in HL trunk insufficient
  /// </summary>
  public class ImpactCalculator {
    /// <summary>
    /// Calculate the impact of a variable within a regression solution.
    /// The impact is calculated by shuffling the variable values (distribution is unchanged but relation to y is broken up).
    /// To account for correlations between input variables the method also shuffles sets of varibles up to the specified maximum number of interactions.
    /// </summary>
    /// <returns>The loss in R² on the specified partition when the variable values are shuffled.</returns>
    public static double Calculate(IRegressionSolution solution, IntRange partition, string variableName, int maxInteractions = 1) {
      var problemData = solution.ProblemData;
      var ds = problemData.Dataset;
      var model = solution.Model;

      var rows = Enumerable.Range(partition.Start, partition.Size).ToArray();
      var y = ds.GetDoubleValues(problemData.TargetVariable, rows);
      var y_pred = model.GetEstimatedValues(ds, rows);
      var original_r = OnlinePearsonsRCalculator.Calculate(y, y_pred, out OnlineCalculatorError error);
      if (error != OnlineCalculatorError.None) original_r = 0;

      var variableNames = model.VariablesUsedForPrediction.ToList();

      // try all combinations of variables up to the given number of interactions and take the minimum impact
      var best_r2 = 0.0;
      var shuffledRows = problemData.AllIndices.ToList();
      shuffledRows.ShuffleInPlace(new FastRandom(1234), partition.Start, partition.End);      

      for (int setSize = 1; setSize <= maxInteractions && setSize <= variableNames.Count; setSize++) {
        var combinations = HeuristicLab.Common.EnumerableExtensions.Combinations(variableNames, setSize)
                                   .Where(comb => comb.Any(v => v == variableName)); // variable combinations that contain the selected variable
        foreach (var combination in combinations) {
          // create a ds and shuffle all values in the variable set (keeping values of the variable set together)
          var shuffledDs = ((Dataset)ds).ToModifiable();
          foreach(var variable in combination) {
            var originalValues = shuffledDs.GetDoubleValues(variable).ToArray();
            var shuffledValues = new List<double>(originalValues); // clone
            foreach(var row in rows) {
              shuffledValues[row] = originalValues[shuffledRows[row]];
            }
            shuffledDs.ReplaceVariable(variable, shuffledValues);
          }

          y_pred = model.GetEstimatedValues(shuffledDs, rows);
          var r2 = OnlinePearsonsRCalculator.Calculate(y, y_pred, out error);
          if (error != OnlineCalculatorError.None) r2 = 0;
          r2 = r2 * r2; // r^2 

          if (r2 > best_r2) {
            best_r2 = r2;
          }
        }
      }
      return original_r * original_r - best_r2; // impact is loss in R² 
    }
  }
}
